// This file was generated by gir (https://github.com/gtk-rs/gir)
// from
// from gir-files (https://github.com/gtk-rs/gir-files.git)
// DO NOT EDIT

use crate::ffi;
use glib::GStr;

#[doc(alias = "SHUMATE_MAP_SOURCE_MFF_RELIEF")]
pub static MAP_SOURCE_MFF_RELIEF: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::SHUMATE_MAP_SOURCE_MFF_RELIEF) };
#[doc(alias = "SHUMATE_MAP_SOURCE_OSM_CYCLE_MAP")]
pub static MAP_SOURCE_OSM_CYCLE_MAP: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::SHUMATE_MAP_SOURCE_OSM_CYCLE_MAP) };
#[doc(alias = "SHUMATE_MAP_SOURCE_OSM_MAPNIK")]
pub static MAP_SOURCE_OSM_MAPNIK: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::SHUMATE_MAP_SOURCE_OSM_MAPNIK) };
#[doc(alias = "SHUMATE_MAP_SOURCE_OSM_TRANSPORT_MAP")]
pub static MAP_SOURCE_OSM_TRANSPORT_MAP: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::SHUMATE_MAP_SOURCE_OSM_TRANSPORT_MAP) };
#[doc(alias = "SHUMATE_MAP_SOURCE_OWM_CLOUDS")]
pub static MAP_SOURCE_OWM_CLOUDS: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::SHUMATE_MAP_SOURCE_OWM_CLOUDS) };
#[doc(alias = "SHUMATE_MAP_SOURCE_OWM_PRECIPITATION")]
pub static MAP_SOURCE_OWM_PRECIPITATION: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::SHUMATE_MAP_SOURCE_OWM_PRECIPITATION) };
#[doc(alias = "SHUMATE_MAP_SOURCE_OWM_PRESSURE")]
pub static MAP_SOURCE_OWM_PRESSURE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::SHUMATE_MAP_SOURCE_OWM_PRESSURE) };
#[doc(alias = "SHUMATE_MAP_SOURCE_OWM_TEMPERATURE")]
pub static MAP_SOURCE_OWM_TEMPERATURE: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::SHUMATE_MAP_SOURCE_OWM_TEMPERATURE) };
#[doc(alias = "SHUMATE_MAP_SOURCE_OWM_WIND")]
pub static MAP_SOURCE_OWM_WIND: &GStr =
    unsafe { GStr::from_utf8_with_nul_unchecked(ffi::SHUMATE_MAP_SOURCE_OWM_WIND) };
